@@ -1,114 +1,8 @@
 /*
- * DMI decode rev 2.0
+ * DMI Decode
  *
- *      (C) 2000-2002 Alan Cox <alan@redhat.com>
- *      (C) 2002 Jean Delvare <khali@linux-fr>
- *
- *      2001-07-02  Matt Domsch <Matt_Domsch@dell.com>
- *      Additional structures displayed per SMBIOS 2.3.1 spec
- *
- *      13-December-2001 Arjan van de Ven <arjanv@redhat.com>
- *      Fix memory bank type (DMI case 6)
- *
- *      2002-08-03  Mark D. Studebaker <mds@paradyne.com>
- *      Better indent in dump_raw_data
- *      Fix return value of dmi_bus_name
- *      Additional sensor fields decoded
- *      Fix compilation warnings
- *
- *      2002-08-06  Jean Delvare <khali@linux-fr.org>
- *      Reposition file pointer after DMI table display
- *      Disable first RSD PTR checksum (was not correct anyway)
- *      Show actual DMI struct count and occupied size
- *      Check for NULL after malloc
- *      Use SEEK_* constants instead of numeric values
- *      Code optimization (and warning fix) in DMI cases 10 and 14
- *      Add else's to avoid unneeded cascaded if's in main loop
- *      Code optimization in DMI information display
- *      Fix all compilation warnings
- *
- *      2002-08-09  Jean Delvare <khali@linux-fr.org>
- *      Better DMI struct count/size error display
- *      More careful memory access in dmi_table
- *      DMI case 13 (Language) decoded
- *      C++ style comments removed
- *      Commented out code removed
- *      DMI 0.0 case handled
- *      Fix return value of dmi_port_type and dmi_port_connector_type
- *
- *      2002-08-23  Alan Cox <alan@redhat.com>
- *      Make the code pass -Wall -pedantic by fixing a few harmless sign of
- *        pointer mismatches
- *      Correct main() prototype
- *      Check for compilers with wrong type sizes
- *
- *      2002-09-17  Larry Lile <llile@dreamworks.com>
- *      Type 16 & 17 structures displayed per SMBIOS 2.3.1 spec
- *
- *      2002-09-20  Dave Johnson <ddj@cascv.brown.edu>
- *      Fix comparisons in dmi_bus_name
- *      Fix comparison in dmi_processor_type
- *      Fix bitmasking in dmi_onboard_type
- *      Fix return value of dmi_temp_loc
- *
- *      2002-09-28  Jean Delvare <khali@linux-fr.org>
- *      Fix missing coma in dmi_bus_name
- *      Remove unwanted bitmaskings in dmi_mgmt_dev_type, dmi_mgmt_addr_type,
- *        dmi_fan_type, dmi_volt_loc, dmi_temp_loc and dmi_status
- *      Fix DMI table read bug ("dmi: read: Success")
- *      Make the code pass -W again
- *      Fix return value of dmi_card_size
- *
- *      2002-10-05  Jean Delvare <khali@linux-fr.org>
- *      More ACPI decoded
- *      More PNP decoded
- *      More SYSID decoded
- *      PCI Interrupt Routing decoded
- *      BIOS32 Service Directory decoded
- *      Sony system detection (unconfirmed)
- *      Checksums verified whenever possible
- *      Better checks on file read and close
- *      Define VERSION and display version at beginning
- *      More secure decoding (won't run off the table in any case)
- *      Do not try to decode more structures than announced
- *      Fix an off-by-one error that caused the last address being
- *        scanned to be 0x100000, not 0xFFFF0 as it should
- *
- *      2002-10-10  Jean Delvare <khali@linux-fr.org>
- *      Remove extra semicolon at the end of dmi_memory_array_use
- *      Fix compilation warnings
- *      Add missing backslash in DMI case 37
- *      Fix BIOS ROM size (DMI case 0)
- *
- *      2002-10-12  Jean Delvare <khali@linux-fr.org>
- *      Fix maximum cache size and installed size being inverted
- *      Fix typos in port types
- *
- *      2002-10-14  Jean Delvare <khali@linux-fr.org>
- *      Fix typo in dmi_memory_array_location
- *      Replace Kbyte by kB in DMI case 16
- *      Add DDR entry in dmi_memory_device_type
- *      Fix extra s in SYSIS
- *
- *      2002-10-15  Jean Delvare <khali@linux-fr.org>
- *      Fix bad index in DMI case 27 (cooling device)
- *
- *      2002-10-18  Jean Delvare <khali@linux-fr.org>
- *      Complete rewrite
- *      Now complies with SMBIOS specification 2.3.3
- *      Removed all non-DMI stuff
- *
- *      2002-10-21  Jean Delvare <khali@linux-fr.org>
- *      Changed supported log type descriptors display
- *      Code optimization in event log status
- *      Remove extra newline in voltage probe accuracy
- *      Display "OEM-specific" if type is 128 or more
- *      Do not display Strings on dump if there are no strings
- *      Add ASCII-filtering to dmi_string
- *      Convert all dates to ISO 8601
- *
- * Licensed under the GNU Public license. If you want to use it in with
- * another license just ask.
+ *   (C) 2000-2002 Alan Cox <alan@redhat.com>
+ *   (C) 2002-2003 Jean Delvare <khali@linux-fr>
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -131,49 +25,43 @@
  *   are deemed to be part of the source code.
  *
  * Unless specified otherwise, all references are aimed at the "System
- * Management BIOS Reference Specification, Version 2.3.3" document,
- * available from http://www.dmtf.org/standards/bios.php.
+ * Management BIOS Reference Specification, Version 2.3.4" document,
+ * available from http://www.dmtf.org/standards/smbios.
  *
  * Note to contributors:
  * Please reference every value you add or modify, especially if the
  * information does not come from the above mentioned specification.
  *
  * Additional references:
- *	- Intel AP-485 revision 21
+ *	- Intel AP-485 revision 23
  *    "Intel Processor Identification and the CPUID Instruction"
  *    http://developer.intel.com/design/xeon/applnots/241618.htm
- *  - DMTF Master MIF version 020507
+ *  - DMTF Master MIF version 030621
  *    "DMTF approved standard groups"
- *    http://www.dmtf.org/standards/standard_dmi.php
- *
- * Thanks to:
- *   Werner Heuser
- *   Alexandre Duret-Lutz
- *   Xavier Roche
- *   Pam Huntley
- *   Gael Stephan
- *   Sebastian Henschel
+ *    http://www.dmtf.org/standards/dmi
+ *  - IPMI 1.5 revision 1.1
+ *    "Intelligent Platform Management Interface Specification"
+ *    http://developer.intel.com/design/servers/ipmi/spec.htm
  */
 
 #include <sys/types.h>
 #include <sys/stat.h>
+#ifdef USE_MMAP
+#include <sys/mman.h>
+#endif /* USE MMAP */
 #include <stdio.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <string.h>
 #include <stdlib.h>
-#include <errno.h>
 
-#define VERSION "2.0"
+#include "version.h"
+#include "types.h"
+#include "util.h"
 
 static const char *out_of_spec = "<OUT OF SPEC>";
 static const char *bad_index = "<BAD INDEX>";
 
-typedef unsigned char u8;
-typedef unsigned short u16;
-typedef signed short i16;
-typedef unsigned int u32;
-
 /*
  * The specification isn't very clear on endianness problems, so we better
  * have macros for these. It also helps us solve problems on systems that
@@ -252,40 +140,6 @@ struct dmi_header
 #define HANDLE(x) x->handle
 #endif
 
-/*
- * Tools
- */
-
-static int myread(int fd, u8 *buf, size_t count, const char *prefix)
-{
-	ssize_t r=1;
-	size_t r2=0;
-	
-	while(r2!=count && r!=0)
-	{
-		r=read(fd, buf+r2, count-r2);
-		if(r==-1)
-		{
-			if(errno!=EINTR)
-			{
-				close(fd);
-				perror(prefix);
-				return -1;
-			}
-		}
-		else
-			r2+=r;
-	}
-	
-	if(r2!=count)
-	{
-		close(fd);
-		fprintf(stderr, "%s: Unexpected end of file\n", prefix);
-		return -1;
-	}
-	
-	return 0;
-}
 
 /*
  * Type-independant Stuff
@@ -382,7 +236,7 @@ static void dmi_dump(struct dmi_header *h, const char *prefix)
 	int row, i;
 	const char *s;
 	
-	printf("%sHeader And Data:\n", prefix);
+	printf("%sHeader and Data:\n", prefix);
 	for(row=0; row<((h->length-1)>>4)+1; row++)
 	{
 		printf("%s\t", prefix);
@@ -392,7 +246,7 @@ static void dmi_dump(struct dmi_header *h, const char *prefix)
 	}
 
 	if(((u8 *)h)[h->length] || ((u8 *)h)[h->length+1])
-	{	
+	{
 		printf("%sStrings:\n", prefix);
 		i=1;
 		while((s=dmi_string(h, i++))!=bad_index)
@@ -893,9 +747,9 @@ static const char *dmi_processor_family(u8 code)
 		"Weitek",
 		NULL, /* 0x81 */
 		"Itanium",
-		NULL, /* 0x83 */
-		NULL,
-		NULL,
+		"Athlon 64",
+		"Opteron",
+		NULL, /* 0x85 */
 		NULL,
 		NULL,
 		NULL,
@@ -946,9 +800,9 @@ static const char *dmi_processor_family(u8 code)
 		"Xeon MP",
 		"Athlon XP",
 		"Athlon MP",
-		NULL, /* 0xB8 */
-		NULL,
-		NULL,
+		"Itanium 2",
+		"Pentium M",
+		NULL, /* 0xBA */
 		NULL,
 		NULL,
 		NULL,
@@ -1028,7 +882,7 @@ static const char *dmi_processor_family(u8 code)
 
 static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *prefix)
 {
-	/* Intel AP-485 revision 21 */
+	/* Intel AP-485 revision 23, table 5 */
 	static const char *flags[32]={
 		"FPU (Floating-point unit on-chip)", /* 0 */
 		"VME (Virtual mode extension)",
@@ -1061,8 +915,13 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 		"HTT (Hyper-threading technology)",
 		"TM (Thermal monitor supported)",
 		NULL, /* 30 */
-		NULL /* 31 */
+		"SBF (Signal break on FERR)" /* 31 */
 	};
+	/*
+	 * Extra flags are now returned in the ECX register when one calls
+	 * the CPUID instruction. Their means are explained in table 6, but
+	 * DMI doesn't support this yet.
+	 */
 	u32 eax;
 	int cpuid=0;
 
@@ -1079,7 +938,7 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 		/*
 		 * 80386 have a different signature.
 		 */
-		printf("%sSignature: Type %u, Family %u, Major Stepping %u, Minor Stepping %u\n",
+		printf("%sSignature: Type %X, Family %X, Major Stepping %X, Minor Stepping %X\n",
 			prefix, dx>>12, (dx>>8)&0xF, (dx>>4)&0xF, dx&0xF);
 		return;
 	}
@@ -1097,7 +956,9 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 	else if((type>=0x0B && type<=0x13) /* Intel, Cyrix */
 	|| (type>=0x18 && type<=0x1D) || type==0x1F /* AMD */
 	|| (type>=0xB0 && type<=0xB3) /* Intel */
-	|| (type>=0xB5 && type<=0xB7)) /* Intel, AMD */
+	|| (type>=0xB5 && type<=0xB7) /* Intel, AMD */
+	|| (type==0xB9) /* Intel */
+	|| (type==0x83 || type==0x84)) /* AMD 64-bit */
 		cpuid=1;
 	else if(type==0x01)
 	{
@@ -1115,7 +976,7 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 		return;
 	
 	eax=DWORD(p);
-	printf("%sSignature: Type %u, Family %u, Model %u, Stepping %u\n",
+	printf("%sSignature: Type %X, Family %X, Model %X, Stepping %X\n",
 		prefix, (eax>>12)&0x3, ((eax>>16)&0xFF0)+((eax>>8)&0x00F),
 		((eax>>12)&0xF0)+((eax>>4)&0x0F), eax&0xF);
 	if(cpuid)
@@ -1150,9 +1011,13 @@ static void dmi_processor_voltage(u8 code)
 	if(code&0x80)
 		printf(" %.1f V", (float)(code&0x7f)/10);
 	else
+	{
 		for(i=0; i<=2; i++)
 			if(code&(1<<i))
 				printf(" %s", voltage[i]);
+		if(code==0x00)
+			printf(" Unknown");
+	}
 }
 
 static void dmi_processor_frequency(u16 code)
@@ -1199,10 +1064,12 @@ static const char *dmi_processor_upgrade(u8 code)
 		"Slot M",
 		"Socket 423",
 		"Socket A (Socket 462)",
-		"Socket 478" /* 0x0F */
+		"Socket 478",
+		"Socket 754",
+		"Socket 940" /* 0x11 */
 	};
 	
-	if(code>=0x01 && code<=0x0F)
+	if(code>=0x01 && code<=0x11)
 		return upgrade[code-0x01];
 	return out_of_spec;
 }
@@ -1335,7 +1202,7 @@ static void dmi_memory_module_types(u16 code, const char *sep)
 		"FPM",
 		"EDO",
 		"Parity",
-		"EDD",
+		"ECC",
 		"SIMM",
 		"DIMM",
 		"Burst EDO",
@@ -1664,7 +1531,8 @@ static const char *dmi_slot_type(u8 code)
 		"AGP",
 		"AGP 2x",
 		"AGP 4x",
-		"PCI-X" /* 0x12 */
+		"PCI-X",
+		"AGP 8x" /* 0x13 */
 	};
 	static const char *type_0xA0[]={
 		"PC-98/C20", /* 0xA0 */
@@ -1674,7 +1542,7 @@ static const char *dmi_slot_type(u8 code)
 		"PC-98/Card" /* 0xA4 */
 	};
 	
-	if(code>=0x01 && code<=0x12)
+	if(code>=0x01 && code<=0x13)
 		return type[code-0x01];
 	if(code>=0xA0 && code<=0xA4)
 		return type_0xA0[code-0xA0];
@@ -1904,7 +1772,7 @@ static const char *dmi_event_log_method(u8 code)
 		"Indexed I/O, two 8-bit index ports, one 8-bit data port",
 		"Indexed I/O, one 16-bit index port, one 8-bit data port",
 		"Memory-mapped physical 32-bit address",
-		"General-pupose non-volatile data functions" /* 0x04 */
+		"General-purpose non-volatile data functions" /* 0x04 */
 	};
 	
 	if(code<=0x04)
@@ -2767,7 +2635,7 @@ static const char *dmi_system_boot_status(u8 code)
 		return status[code];
 	if(code>=128 && code<=191)
 		return "OEM-specific";
-	if(code>=192 && code<=255)
+	if(code>=192)
 		return "Product-specific";
 	return out_of_spec;
 }
@@ -2875,7 +2743,21 @@ static const char *dmi_ipmi_interface_type(u8 code)
 	};
 	
 	if(code<=0x03)
-		return type[code-0x01];
+		return type[code];
+	return out_of_spec;
+}
+
+static const char *dmi_ipmi_register_spacing(u8 code)
+{
+	/* IPMI 1.5 */
+	static const char *spacing[]={
+		"Successive Byte Boundaries", /* 0x00 */
+		"32-bit Boundaries",
+		"16-byte Boundaries" /* 0x02 */
+	};
+	
+	if(code<=0x02)
+		return spacing[code];
 	return out_of_spec;
 }
 
@@ -3762,19 +3644,48 @@ static void dmi_decode(u8 *data, u16 ver)
 			break;
 		
 		case 38: /* 3.3.39 IPMI Device Information */
+			/*
+			 * We use the word "Version" instead of "Revision", conforming to
+			 * the IPMI 1.5 specification.
+			 */
 			printf("\tIPMI Device Information\n");
 			if(h->length<0x10) break;
 			printf("\t\tInterface Type: %s\n",
 				dmi_ipmi_interface_type(data[0x04]));
-			printf("\t\tSpecification Revision: %u.%u\n",
+			printf("\t\tSpecification Version: %u.%u\n",
 				data[0x05]>>4, data[0x05]&0x0F);
-			printf("\t\tI2C Slave Address: %u\n",
-				data[0x06]);
-			if(data[0x07]==0xFF)
+			printf("\t\tI2C Slave Address: 0x%02x\n",
+				data[0x06]>>1);
+			if(data[0x07]!=0xFF)
 				printf("\t\tNV Storage Device Address: %u\n",
 					data[0x07]);
-			printf("\t\tBase Address: 0x%08X (%s)\n",
-				DWORD(data+0x08), DWORD(data+0x08)&1?"I/O":"memory-mapped");
+			else
+				printf("\t\tNV Storage Device: Not Present\n");
+			if(h->length<0x12)
+			{
+				printf("\t\tBase Address: 0x%08X%08X (%s)\n",
+					QWORD(data+0x08).h, QWORD(data+0x08).l,
+					QWORD(data+0x08).l&1?"I/O":"Memory-mapped");
+				break;
+			}
+			printf("\t\tBase Address: 0x%08X%08X (%s)\n",
+				QWORD(data+0x08).h,
+				(QWORD(data+0x08).l&~1)|((data[0x10]>>5)&1),
+				QWORD(data+0x08).l&1?"I/O":"Memory-mapped");
+			printf("\t\tRegister Spacing: %s\n",
+				dmi_ipmi_register_spacing(data[0x10]>>6));
+			if(data[0x10]&(1<<3))
+			{
+				printf("\t\tInterrupt Polarity: %s\n",
+					data[0x10]&(1<<1)?"Active High":"Active Low");
+				printf("\t\tInterrupt Trigger Mode: %s\n",
+					data[0x10]&(1<<0)?"Level":"Edge");
+			}
+			if(data[0x11]!=0x00)
+			{
+				printf("\t\tInterrupt Number: %x\n",
+					data[0x11]);
+			}
 			break;
 		
 		case 39: /* 3.3.40 System Power Supply */
@@ -3847,19 +3758,40 @@ static void dmi_table(int fd, u32 base, u16 len, u16 num, u16 ver, const char *p
 	u8 *buf;
 	u8 *data;
 	int i=0;
+#ifdef USE_MMAP
+	u32 mmoffset;
+	void *mmp;
+#endif /* USE_MMAP */
 	
 	printf("%u structures occupying %u bytes.\n",
 		num, len);
 	printf("Table at 0x%08X.\n",
 		base);
 	
-	if(base>0xFFFFF) return;
-	
 	if((buf=malloc(len))==NULL)
 	{
 		perror(pname);
 		return;
 	}
+#ifdef USE_MMAP
+	mmoffset=base%getpagesize();
+	/*
+	 * We were previously using PROT_WRITE and MAP_PRIVATE, but it caused
+	 * trouble once. So we are now mapping in read-only mode and copying
+	 * the interesting block into a regular memory buffer (similar to what
+	 * we do when not using mmap.)
+	 */
+	mmp=mmap(0, mmoffset+len, PROT_READ, MAP_SHARED, fd, base-mmoffset);
+	if(mmp==MAP_FAILED)
+	{
+		free(buf);
+    	perror(devmem);
+    	return;
+	}
+	memcpy(buf, (u8 *)mmp+mmoffset, len);
+	if(munmap(mmp, mmoffset+len)==-1)
+		perror(devmem);
+#else /* USE_MMAP */
 	if(lseek(fd, (off_t)base, SEEK_SET)==-1)
 	{
 		perror(devmem);
@@ -3868,8 +3800,12 @@ static void dmi_table(int fd, u32 base, u16 len, u16 num, u16 ver, const char *p
 	if(myread(fd, buf, len, devmem)==-1)
 	{
 		free(buf);
+		printf("Table is unreachable, sorry. Try compiling dmidecode with -DUSE_MMAP.\n"
+			"This problem is known on the IBM T23, T30 and X30 laptops, the Fujitsu-Siemens\n"
+			"S-4582 laptop as well as IA-64 systems. If your system differ, please report\n");
 		exit(1);
 	}
+#endif /* USE_MMAP */
 	
 	data=buf;
 	while(i<num && data+sizeof(struct dmi_header)<=buf+len)
@@ -3887,6 +3823,8 @@ static void dmi_table(int fd, u32 base, u16 len, u16 num, u16 ver, const char *p
 		next+=2;
 		if(next-buf<=len)
 			dmi_decode(data, ver);
+		else
+			printf("\t<TRUNCATED>\n");
 		
 		data=next;
 		i++;
@@ -3896,28 +3834,46 @@ static void dmi_table(int fd, u32 base, u16 len, u16 num, u16 ver, const char *p
 		printf("Wrong DMI structures count: %d announced, only %d decoded.\n",
 			num, i);
 	if(data-buf!=len)
-		printf("Wrong DMI structures length: %d bytes announced, %d bytes decoded.\n",
-			len, data-buf);
+		printf("Wrong DMI structures length: %d bytes announced, structures occupy %d bytes.\n",
+			len, (unsigned int)(data-buf));
 	
 	free(buf);
 }
 
 
-static int checksum(u8 *buf, u8 len)
+static int smbios_decode(u8 *buf, int fd, const char *pname, const char *devmem)
 {
-	u8 sum=0, a;
+	if(checksum(buf, buf[0x05])
+	 && memcmp(buf+0x10, "_DMI_", 5)==0
+	 && checksum(buf+0x10, 0x0F))
+	{
+		printf("SMBIOS %u.%u present.\n",
+			buf[0x06], buf[0x07]);
+		dmi_table(fd, DWORD(buf+0x18), WORD(buf+0x16), WORD(buf+0x1C),
+			(buf[0x06]<<8)+buf[0x07], pname, devmem);
+		return 1;
+	}
 	
-	for(a=0; a<len; a++)
-		sum+=buf[a];
-	return (sum==0);
+	return 0;
 }
 
 int main(int argc, const char *argv[])
 {
-	u8 buf[0x20];
-	int fd;
+	int fd, found=0;
 	off_t fp=0xF0000;
 	const char *devmem="/dev/mem";
+#ifdef __IA64__
+	FILE *efi_systab;
+	char linebuf[64];
+#ifdef USE_MMAP
+	u32 mmoffset;
+	void *mmp;
+#else /* USE_MMAP */
+	u8 buf[0x20];
+#endif /* USE_MMAP */
+#else /* __IA64__ */
+	u8 buf[0x20];
+#endif /* __IA64__ */
 	
 	if(sizeof(u8)!=1 || sizeof(u16)!=2 || sizeof(u32)!=4 || '\0'!=0)
 	{
@@ -3927,36 +3883,83 @@ int main(int argc, const char *argv[])
 	
 	if(argc>=2)
 		devmem=argv[1];
-	if((fd=open(devmem, O_RDONLY))==-1 || lseek(fd, fp, SEEK_SET)==-1)
+	if((fd=open(devmem, O_RDONLY))==-1)
 	{
 		perror(devmem);
 		exit(1);
 	}
 	
 	printf("# dmidecode %s\n", VERSION);
-	while(fp<0xFFFFF)
+	
+#ifdef __IA64__
+	if((efi_systab=fopen("/proc/efi/systab", "r"))==NULL)
+	{
+		perror("/proc/efi/systab");
+		exit(1);
+	}
+	while((fgets(linebuf, sizeof(linebuf)-1, efi_systab))!=NULL)
+	{
+		char* addr=memchr(linebuf, '=', strlen(linebuf));
+		*(addr++)='\0';
+		if(strcmp(linebuf, "SMBIOS")==0)
+		{
+			fp=strtol(addr, NULL, 0);
+			printf("# SMBIOS entry point at 0x%08lx\n", fp);
+		}
+	}
+	if(fclose(efi_systab)!=0)
+		perror("/proc/efi/systab");
+
+#ifdef USE_MMAP
+	mmoffset=fp%getpagesize();
+	mmp=mmap(0, mmoffset+0x20, PROT_READ, MAP_PRIVATE, fd, fp-mmoffset);
+    if(mmp==MAP_FAILED)
+    {
+       perror(devmem);
+       exit(1);
+    }
+
+	smbios_decode(((u8 *)mmp)+mmoffset, fd, argv[0], devmem);
+
+	if(munmap(mmp, mmoffset+0x20)==-1)
+		perror(devmem);
+#else /* USE_MMAP */
+	if(lseek(fd, fp, SEEK_SET)==-1)
+	{
+		perror(devmem);
+		exit(1);
+	}
+	if(myread(fd, buf, 0x20, devmem)==-1)
+		exit(1);
+
+	smbios_decode(buf, fd, argv[0], devmem);
+#endif /* USE_MMAP */
+	found++;
+#else /* __IA64__ */
+	if(lseek(fd, fp, SEEK_SET)==-1)
+	{
+		perror(devmem);
+		exit(1);
+	}
+	while(fp<=0xFFFF0)
 	{
 		if(myread(fd, buf, 0x10, devmem)==-1)
 			exit(1);
 		fp+=16;
 		
-		if(memcmp(buf, "_SM_", 4)==0 && fp<0xFFFFF)
+		if(memcmp(buf, "_SM_", 4)==0 && fp<=0xFFFF0)
 		{
 			if(myread(fd, buf+0x10, 0x10, devmem)==-1)
 				exit(1);
 			fp+=16;
 			
-			if(checksum(buf, buf[0x05])
-			 && memcmp(buf+0x10, "_DMI_", 5)==0
-			 && checksum(buf+0x10, 0x0F))
+			if(smbios_decode(buf, fd, argv[0], devmem))
 			{
-				printf("SMBIOS %u.%u present.\n",
-					buf[0x06], buf[0x07]);
-				dmi_table(fd, DWORD(buf+0x18), WORD(buf+0x16), WORD(buf+0x1C),
-					(buf[0x06]<<8)+buf[0x07], argv[0], devmem);
-				
+#ifndef USE_MMAP
 				/* dmi_table moved us far away */
 				lseek(fd, fp, SEEK_SET);
+#endif /* USE_MMAP */
+				found++;
 			}
 		}
 		else if(memcmp(buf, "_DMI_", 5)==0
@@ -3967,10 +3970,14 @@ int main(int argc, const char *argv[])
 			dmi_table(fd, DWORD(buf+0x08), WORD(buf+0x06), WORD(buf+0x0C),
 				((buf[0x0E]&0xF0)<<4)+(buf[0x0E]&0x0F), argv[0], devmem);
 			
+#ifndef USE_MMAP
 			/* dmi_table moved us far away */
 			lseek(fd, fp, SEEK_SET);
+#endif /* USE_MMAP */
+			found++;
 		}
 	}
+#endif /* __IA64__ */
 	
 	if(close(fd)==-1)
 	{
@@ -3978,5 +3985,8 @@ int main(int argc, const char *argv[])
 		exit(1);
 	}
 	
+	if(!found)
+		printf("# No SMBIOS nor DMI entry point found, sorry.\n");
+
 	return 0;
 }
